@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -31,9 +31,7 @@
 #include <errno.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/hci.h>
 #include <bluetooth/bnep.h>
-#include <bluetooth/l2cap.h>
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
 #include <netinet/in.h>
@@ -44,67 +42,58 @@
 #include "../src/dbus-common.h"
 #include "../src/adapter.h"
 
-#include "logging.h"
+#include "log.h"
 #include "error.h"
 #include "sdpd.h"
+#include "btio.h"
 #include "glib-helper.h"
 
-#include "bridge.h"
 #include "common.h"
+#include "server.h"
 
-#define NETWORK_PEER_INTERFACE "org.bluez.NetworkPeer"
-#define NETWORK_HUB_INTERFACE "org.bluez.NetworkHub"
-#define NETWORK_ROUTER_INTERFACE "org.bluez.NetworkRouter"
+#define NETWORK_SERVER_INTERFACE "org.bluez.NetworkServer"
 #define SETUP_TIMEOUT		1
 
 /* Pending Authorization */
-struct setup_session {
-	char		*address;	/* Remote Bluetooth Address */
-	uint16_t	dst_role;	/* Destination role */
-	uint16_t	src_role;	/* Source role */
-	int		nsk;		/* L2CAP socket */
+struct network_session {
+	bdaddr_t	dst;		/* Remote Bluetooth Address */
+	GIOChannel	*io;		/* Pending connect channel */
 	guint		watch;		/* BNEP socket watch */
 };
 
-struct timeout {
-	guint	id;		/* Timeout id */
-	guint	watch;		/* BNEP socket watch */
-};
-
 struct network_adapter {
-	bdaddr_t	src;		/* Bluetooth Local Address */
-	char		*path;		/* D-Bus path */
+	struct btd_adapter *adapter;	/* Adapter pointer */
 	GIOChannel	*io;		/* Bnep socket */
-	struct timeout	*to;		/* Socket timeout */
+	struct network_session *setup;	/* Setup in progress */
 	GSList		*servers;	/* Server register to adapter */
 };
 
 /* Main server structure */
 struct network_server {
+	bdaddr_t	src;		/* Bluetooth Local Address */
 	char		*iface;		/* DBus interface */
 	char		*name;		/* Server service name */
-	char		*range;		/* IP Address range */
-	gboolean	enable;		/* Enable flag */
+	char		*bridge;	/* Bridge name */
 	uint32_t	record_id;	/* Service record id */
 	uint16_t	id;		/* Service class identifier */
-	GSList		*clients;	/* Active connections */
+	GSList		*sessions;	/* Active connections */
 	struct network_adapter *na;	/* Adapter reference */
+	guint		watch_id;	/* Client service watch */
 };
 
 static DBusConnection *connection = NULL;
-static struct setup_session *setup = NULL;
 static GSList *adapters = NULL;
-static const char *prefix = NULL;
 static gboolean security = TRUE;
 
-static struct network_adapter *find_adapter(GSList *list, const char *path)
+static struct network_adapter *find_adapter(GSList *list,
+					struct btd_adapter *adapter)
 {
 	GSList *l;
 
 	for (l = list; l; l = l->next) {
 		struct network_adapter *na = l->data;
 
-		if (g_str_equal(na->path, path))
+		if (na->adapter == adapter)
 			return na;
 	}
 
@@ -125,28 +114,6 @@ static struct network_server *find_server(GSList *list, uint16_t id)
 	return NULL;
 }
 
-static struct setup_session *setup_session_new(gchar *address,
-		uint16_t dst_role, uint16_t src_role, int nsk, guint watch)
-{
-	struct setup_session *setup;
-
-	setup = g_new0(struct setup_session, 1);
-	setup->address = g_strdup(address);
-	setup->dst_role = dst_role;
-	setup->src_role = src_role;
-	setup->nsk = nsk;
-	setup->watch = watch;
-
-	return setup;
-}
-
-static void setup_session_free(struct setup_session *setup)
-{
-	g_source_remove(setup->watch);
-	g_free(setup->address);
-	g_free(setup);
-}
-
 static void add_lang_attr(sdp_record_t *r)
 {
 	sdp_lang_attr_t base_lang;
@@ -172,7 +139,7 @@ static sdp_record_t *server_record_new(const char *name, uint16_t id)
 	uint16_t security_desc = (security ? 0x0001 : 0x0000);
 	uint16_t net_access_type = 0xfffe;
 	uint32_t max_net_access_rate = 0;
-	const char *desc = "BlueZ PAN service";
+	const char *desc = "Network service";
 	sdp_record_t *record;
 
 	record = sdp_record_alloc();
@@ -225,6 +192,7 @@ static sdp_record_t *server_record_new(const char *name, uint16_t id)
 		sdp_set_info_attr(record, name, NULL, desc);
 		break;
 	default:
+		sdp_record_free(record);
 		return NULL;
 	}
 
@@ -297,90 +265,37 @@ static ssize_t send_bnep_ctrl_rsp(int sk, uint16_t val)
 	return send(sk, &rsp, sizeof(rsp), 0);
 }
 
-static int server_connadd(struct network_server *ns, int nsk,
-			const gchar *address, uint16_t dst_role)
+static int server_connadd(struct network_server *ns,
+				struct network_session *session,
+				uint16_t dst_role)
 {
 	char devname[16];
-	const char *bridge;
-	int err;
+	int err, nsk;
 
-	/* Server can be disabled in the meantime */
-	if (ns->enable == FALSE)
-		return -EPERM;
-
-	memset(devname, 0, 16);
-	strncpy(devname, prefix, strlen(prefix));
+	memset(devname, 0, sizeof(devname));
+	strcpy(devname, "bnep%d");
 
+	nsk = g_io_channel_unix_get_fd(session->io);
 	err = bnep_connadd(nsk, dst_role, devname);
 	if (err < 0)
 		return err;
 
 	info("Added new connection: %s", devname);
 
-	bridge = bridge_get_name(ns->id);
-	if (bridge) {
-		if (bridge_add_interface(ns->id, devname) < 0) {
-			error("Can't add %s to the bridge %s: %s(%d)",
-					devname, bridge, strerror(errno),
-					errno);
-			return -EPERM;
-		}
+	if (bnep_add_to_bridge(devname, ns->bridge) < 0) {
+		error("Can't add %s to the bridge %s: %s(%d)",
+				devname, ns->bridge, strerror(errno), errno);
+		return -EPERM;
+	}
 
-		bnep_if_up(devname, 0);
-	} else
-		bnep_if_up(devname, ns->id);
+	bnep_if_up(devname);
 
-	ns->clients = g_slist_append(ns->clients, g_strdup(address));
+	ns->sessions = g_slist_append(ns->sessions, session);
 
 	return 0;
 }
 
-static void req_auth_cb(DBusError *derr, void *user_data)
-{
-	struct network_server *ns = user_data;
-	uint16_t val;
-
-	if (!setup) {
-		info("Authorization cancelled: Client exited");
-		return;
-	}
-
-	if (derr) {
-		error("Access denied: %s", derr->message);
-
-		val = BNEP_CONN_NOT_ALLOWED;
-		goto done;
-	}
-
-	if (server_connadd(ns, setup->nsk,
-			setup->address, setup->dst_role) < 0)
-		val = BNEP_CONN_NOT_ALLOWED;
-	else
-		val = BNEP_SUCCESS;
-
-done:
-	send_bnep_ctrl_rsp(setup->nsk, val);
-	setup_session_free(setup);
-	setup = NULL;
-}
-
-static int authorize_connection(struct network_server *ns, const char *address)
-{
-	struct network_adapter *na = ns->na;
-	const char *uuid;
-	bdaddr_t dst;
-	int ret_val;
-
-	uuid = bnep_uuid(ns->id);
-	str2ba(address, &dst);
-
-	ret_val = btd_request_authorization(&na->src, &dst, uuid,
-				req_auth_cb, ns);
-
-	return ret_val;
-}
-
-static uint16_t inline bnep_setup_chk(uint16_t dst_role, uint16_t src_role)
+static uint16_t bnep_setup_chk(uint16_t dst_role, uint16_t src_role)
 {
 	/* Allowed PAN Profile scenarios */
 	switch (dst_role) {
@@ -391,8 +306,8 @@ static uint16_t inline bnep_setup_chk(uint16_t dst_role, uint16_t src_role)
 		return BNEP_CONN_INVALID_SRC;
 	case BNEP_SVC_PANU:
 		if (src_role == BNEP_SVC_PANU ||
-			src_role == BNEP_SVC_GN ||
-			src_role == BNEP_SVC_NAP)
+				src_role == BNEP_SVC_GN ||
+				src_role == BNEP_SVC_NAP)
 			return 0;
 
 		return BNEP_CONN_INVALID_SRC;
@@ -426,18 +341,40 @@ static uint16_t bnep_setup_decode(struct bnep_setup_conn_req *req,
 	return 0;
 }
 
+static void session_free(void *data)
+{
+	struct network_session *session = data;
+
+	if (session->watch)
+		g_source_remove(session->watch);
+
+	if (session->io)
+		g_io_channel_unref(session->io);
+
+	g_free(session);
+}
+
+static void setup_destroy(void *user_data)
+{
+	struct network_adapter *na = user_data;
+	struct network_session *setup = na->setup;
+
+	if (!setup)
+		return;
+
+	na->setup = NULL;
+
+	session_free(setup);
+}
+
 static gboolean bnep_setup(GIOChannel *chan,
 			GIOCondition cond, gpointer user_data)
 {
 	struct network_adapter *na = user_data;
-	struct timeout *to = na->to;
 	struct network_server *ns;
 	uint8_t packet[BNEP_MTU];
 	struct bnep_setup_conn_req *req = (void *) packet;
-	struct sockaddr_l2 sa;
-	socklen_t size;
-	char address[18];
-	uint16_t rsp, src_role, dst_role;
+	uint16_t src_role, dst_role, rsp = BNEP_CONN_NOT_ALLOWED;
 	int n, sk;
 
 	if (cond & G_IO_NVAL)
@@ -468,47 +405,28 @@ static gboolean bnep_setup(GIOChannel *chan,
 	if (rsp)
 		goto reply;
 
-	size = sizeof(sa);
-	if (getsockname(sk, (struct sockaddr *) &sa, &size) < 0) {
-		rsp = BNEP_CONN_NOT_ALLOWED;
-		goto reply;
-	}
-
-	ba2str(&sa.l2_bdaddr, address);
 	ns = find_server(na->servers, dst_role);
-	if (!ns || ns->enable == FALSE) {
-		error("Server unavailable: %s (0x%x)", address, dst_role);
-		rsp = BNEP_CONN_NOT_ALLOWED;
+	if (!ns) {
+		error("Server unavailable: (0x%x)", dst_role);
 		goto reply;
 	}
 
-	if (getpeername(sk, (struct sockaddr *) &sa, &size) < 0) {
-		rsp = BNEP_CONN_NOT_ALLOWED;
+	if (!ns->record_id) {
+		error("Service record not available");
 		goto reply;
 	}
 
-	ba2str(&sa.l2_bdaddr, address);
-
-	if (setup) {
-		error("Connection rejected: Pending authorization");
-		rsp = BNEP_CONN_NOT_ALLOWED;
+	if (!ns->bridge) {
+		error("Bridge interface not configured");
 		goto reply;
 	}
 
-	setup = setup_session_new(address, dst_role, src_role, sk, to->watch);
-
-	/* Wait authorization before reply success */
-	if (authorize_connection(ns, address) < 0) {
-		setup_session_free(setup);
-		setup = NULL;
-		rsp = BNEP_CONN_NOT_ALLOWED;
+	if (server_connadd(ns, na->setup, dst_role) < 0)
 		goto reply;
-	}
 
-	g_source_remove(to->id);
-	to->id = 0;
+	na->setup = NULL;
 
-	return TRUE;
+	rsp = BNEP_SUCCESS;
 
 reply:
 	send_bnep_ctrl_rsp(sk, rsp);
@@ -516,79 +434,119 @@ reply:
 	return FALSE;
 }
 
-static void setup_destroy(void *user_data)
-{
-	struct network_adapter *na = user_data;
-	struct timeout *to = na->to;
-
-	if (to->id)
-		g_source_remove(to->id);
-
-	g_free(to);
-}
-
-static gboolean timeout_cb(void *user_data)
-{
-	struct timeout *to = user_data;
-
-	to->id = 0;
-	g_source_remove(to->watch);
-
-	return FALSE;
-}
-
-static void connect_event(GIOChannel *chan, int err, const bdaddr_t *src,
-				const bdaddr_t *dst, gpointer user_data)
+static void connect_event(GIOChannel *chan, GError *err, gpointer user_data)
 {
 	struct network_adapter *na = user_data;
 
-	if (err < 0) {
-		error("accept(): %s(%d)", strerror(errno), errno);
+	if (err) {
+		error("%s", err->message);
+		setup_destroy(na);
 		return;
 	}
 
 	g_io_channel_set_close_on_unref(chan, TRUE);
 
-	/*
-	 * BNEP_SETUP_CONNECTION_REQUEST_MSG shall be received and
-	 * user shall authorize the incomming connection before
-	 * the time expires.
-	 */
-	na->to = g_malloc0(sizeof(struct timeout));
-	na->to->id = g_timeout_add_seconds(SETUP_TIMEOUT, timeout_cb, na->to);
-	na->to->watch = g_io_add_watch_full(chan, G_PRIORITY_DEFAULT,
+	na->setup->watch = g_io_add_watch_full(chan, G_PRIORITY_DEFAULT,
 				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
 				bnep_setup, na, setup_destroy);
-	g_io_channel_unref(chan);
-
-	return;
 }
 
-int server_init(DBusConnection *conn, const char *iface_prefix,
-		gboolean secure)
+static void auth_cb(DBusError *derr, void *user_data)
+{
+	struct network_adapter *na = user_data;
+	GError *err = NULL;
+
+	if (derr) {
+		error("Access denied: %s", derr->message);
+		goto reject;
+	}
+
+	if (!bt_io_accept(na->setup->io, connect_event, na, NULL,
+							&err)) {
+		error("bt_io_accept: %s", err->message);
+		g_error_free(err);
+		goto reject;
+	}
+
+	return;
+
+reject:
+	g_io_channel_shutdown(na->setup->io, TRUE, NULL);
+	setup_destroy(na);
+}
+
+static void confirm_event(GIOChannel *chan, gpointer user_data)
+{
+	struct network_adapter *na = user_data;
+	struct network_server *ns;
+	int perr;
+	bdaddr_t src, dst;
+	char address[18];
+	GError *err = NULL;
+
+	bt_io_get(chan, BT_IO_L2CAP, &err,
+			BT_IO_OPT_SOURCE_BDADDR, &src,
+			BT_IO_OPT_DEST_BDADDR, &dst,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		goto drop;
+	}
+
+	DBG("BNEP: incoming connect from %s", address);
+
+	if (na->setup) {
+		error("Refusing connect from %s: setup in progress", address);
+		goto drop;
+	}
+
+	ns = find_server(na->servers, BNEP_SVC_NAP);
+	if (!ns)
+		goto drop;
+
+	if (!ns->record_id)
+		goto drop;
+
+	if (!ns->bridge)
+		goto drop;
+
+	na->setup = g_new0(struct network_session, 1);
+	bacpy(&na->setup->dst, &dst);
+	na->setup->io = g_io_channel_ref(chan);
+
+	perr = btd_request_authorization(&src, &dst, BNEP_SVC_UUID,
+					auth_cb, na);
+	if (perr < 0) {
+		error("Refusing connect from %s: %s (%d)", address,
+				strerror(-perr), -perr);
+		setup_destroy(na);
+		goto drop;
+	}
+
+	return;
+
+drop:
+	g_io_channel_shutdown(chan, TRUE, NULL);
+}
+
+int server_init(DBusConnection *conn, gboolean secure)
 {
 	security = secure;
 	connection = dbus_connection_ref(conn);
-	prefix = iface_prefix;
-
-	if (bridge_create(BNEP_SVC_GN) < 0)
-		error("Can't create GN bridge");
 
 	return 0;
 }
 
-void server_exit()
+void server_exit(void)
 {
-	if (bridge_remove(BNEP_SVC_GN) < 0)
-		error("Can't remove GN bridge");
-
 	dbus_connection_unref(connection);
 	connection = NULL;
 }
 
 static uint32_t register_server_record(struct network_server *ns)
 {
-	struct network_adapter *na = ns->na;
 	sdp_record_t *record;
 
 	record = server_record_new(ns->name, ns->id);
@@ -597,207 +555,101 @@ static uint32_t register_server_record(struct network_server *ns)
 		return 0;
 	}
 
-	if (add_record_to_server(&na->src, record) < 0) {
+	if (add_record_to_server(&ns->src, record) < 0) {
 		error("Failed to register service record");
 		sdp_record_free(record);
 		return 0;
 	}
 
-	debug("register_server_record: got record id 0x%x", record->handle);
+	DBG("got record id 0x%x", record->handle);
 
 	return record->handle;
 }
 
-
-static inline DBusMessage *failed(DBusMessage *msg, const char *description)
+static void server_disconnect(DBusConnection *conn, void *user_data)
 {
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-				description);
-}
+	struct network_server *ns = user_data;
 
-static inline DBusMessage *invalid_arguments(DBusMessage *msg,
-					const char *description)
-{
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
-				description);
-}
+	ns->watch_id = 0;
 
-static DBusMessage *enable(DBusConnection *conn,
-			DBusMessage *msg, void *data)
-{
-	struct network_server *ns = data;
-	DBusMessage *reply;
-
-	if (ns->enable)
-		return g_dbus_create_error(msg, ERROR_INTERFACE
-						".AlreadyExist",
-						"Server already enabled");
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
-	/* Add the service record */
-	ns->record_id = register_server_record(ns);
-	if (!ns->record_id) {
-		dbus_message_unref(reply);
-		return failed(msg, "Service record registration failed");
-	}
-
-	ns->enable = TRUE;
-
-	return reply;
-}
-
-static void kill_connection(void *data, void *udata)
-{
-	const char *address = data;
-	bdaddr_t dst;
-
-	str2ba(address, &dst);
-	bnep_kill_connection(&dst);
-}
-
-static DBusMessage *disable(DBusConnection *conn,
-				DBusMessage *msg, void *data)
-{
-	struct network_server *ns = data;
-	DBusMessage *reply;
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
-	if (!ns->enable)
-		return failed(msg, "Not enabled");
-
-	/* Remove the service record */
 	if (ns->record_id) {
 		remove_record_from_server(ns->record_id);
 		ns->record_id = 0;
 	}
 
-	ns->enable = FALSE;
-
-	g_slist_foreach(ns->clients, (GFunc) kill_connection, NULL);
-
-	return reply;
+	g_free(ns->bridge);
+	ns->bridge = NULL;
 }
 
-static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
-				const char *name, void *data)
-{
-	struct network_server *ns = data;
-	DBusMessage *reply;
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
-
-	if (!name || (strlen(name) == 0))
-		return invalid_arguments(msg, "Invalid name");
-
-	if (ns->name)
-		g_free(ns->name);
-	ns->name = g_strdup(name);
-
-	if (ns->enable && ns->record_id) {
-		uint32_t handle = register_server_record(ns);
-		if (!handle) {
-			dbus_message_unref(reply);
-			return failed(msg,
-				"Service record attribute update failed");
-		}
-
-		remove_record_from_server(ns->record_id);
-		ns->record_id = handle;
-	}
-
-	return reply;
-}
-
-static DBusMessage *get_properties(DBusConnection *conn,
+static DBusMessage *register_server(DBusConnection *conn,
 				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
-	DBusMessageIter iter;
-	DBusMessageIter dict;
-	const char *uuid;
+	const char *uuid, *bridge;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &uuid,
+				DBUS_TYPE_STRING, &bridge, DBUS_TYPE_INVALID))
+		return NULL;
+
+	if (g_strcmp0(uuid, "nap"))
+		return btd_error_failed(msg, "Invalid UUID");
+
+	if (ns->record_id)
+		return btd_error_already_exists(msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
 		return NULL;
 
-	dbus_message_iter_init_append(reply, &iter);
+	ns->record_id = register_server_record(ns);
+	if (!ns->record_id)
+		return btd_error_failed(msg, "SDP record registration failed");
 
-	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
-			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
-			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
-			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+	g_free(ns->bridge);
+	ns->bridge = g_strdup(bridge);
 
-	dict_append_entry(&dict, "Name", DBUS_TYPE_STRING, &ns->name);
-
-	uuid = bnep_uuid(ns->id);
-	dict_append_entry(&dict, "Uuid", DBUS_TYPE_STRING, &uuid);
-
-	dict_append_entry(&dict, "Enabled", DBUS_TYPE_BOOLEAN, &ns->enable);
-
-	dbus_message_iter_close_container(&iter, &dict);
+	ns->watch_id = g_dbus_add_disconnect_watch(conn,
+					dbus_message_get_sender(msg),
+					server_disconnect, ns, NULL);
 
 	return reply;
 }
 
-static DBusMessage *set_property(DBusConnection *conn,
+static DBusMessage *unregister_server(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	DBusMessageIter iter;
-	DBusMessageIter sub;
-	const char *property;
+	struct network_server *ns = data;
+	DBusMessage *reply;
+	const char *uuid;
 
-	if (!dbus_message_iter_init(msg, &iter))
-		return invalid_arguments(msg, "Not a dict");
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &uuid,
+							DBUS_TYPE_INVALID))
+		return NULL;
 
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
-		return invalid_arguments(msg, "Key not a string");
+	if (g_strcmp0(uuid, "nap"))
+		return btd_error_failed(msg, "Invalid UUID");
 
-	dbus_message_iter_get_basic(&iter, &property);
-	dbus_message_iter_next(&iter);
+	reply = dbus_message_new_method_return(msg);
+	if (!reply)
+		return NULL;
 
-	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
-		return invalid_arguments(msg, "Value not a variant");
-	dbus_message_iter_recurse(&iter, &sub);
+	g_dbus_remove_watch(conn, ns->watch_id);
 
-	if (g_str_equal("Name", property)) {
-		const char *name;
+	server_disconnect(conn, ns);
 
-		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_STRING)
-			return invalid_arguments(msg, "Value not string");
-		dbus_message_iter_get_basic(&sub, &name);
-
-		return set_name(conn, msg, name, data);
-	} else if (g_str_equal("Enabled", property)) {
-		gboolean enabled;
-
-		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
-			return invalid_arguments(msg, "Value not boolean");
-		dbus_message_iter_get_basic(&sub, &enabled);
-
-		return enabled ? enable(conn, msg, data) :
-				disable(conn, msg, data);
-	}
-
-	return invalid_arguments(msg, "Property does not exist");
+	return reply;
 }
 
 static void adapter_free(struct network_adapter *na)
 {
 	if (na->io != NULL) {
-		g_io_channel_close(na->io);
+		g_io_channel_shutdown(na->io, TRUE, NULL);
 		g_io_channel_unref(na->io);
 	}
 
-	g_free(na->path);
+	setup_destroy(na);
+	btd_adapter_unref(na->adapter);
 	g_free(na);
 }
 
@@ -810,18 +662,13 @@ static void server_free(struct network_server *ns)
 	if (ns->record_id)
 		remove_record_from_server(ns->record_id);
 
-	if (ns->iface)
-		g_free(ns->iface);
+	g_free(ns->iface);
+	g_free(ns->name);
+	g_free(ns->bridge);
 
-	if (ns->name)
-		g_free(ns->name);
-
-	if (ns->range)
-		g_free(ns->range);
-
-	if (ns->clients) {
-		g_slist_foreach(ns->clients, (GFunc) g_free, NULL);
-		g_slist_free(ns->clients);
+	if (ns->sessions) {
+		g_slist_foreach(ns->sessions, (GFunc) session_free, NULL);
+		g_slist_free(ns->sessions);
 	}
 
 	g_free(ns);
@@ -832,8 +679,8 @@ static void path_unregister(void *data)
 	struct network_server *ns = data;
 	struct network_adapter *na = ns->na;
 
-	info("Unregistered interface %s on path %s",
-		ns->iface, na->path);
+	DBG("Unregistered interface %s on path %s",
+		ns->iface, adapter_get_path(na->adapter));
 
 	na->servers = g_slist_remove(na->servers, ns);
 	server_free(ns);
@@ -846,76 +693,68 @@ static void path_unregister(void *data)
 }
 
 static GDBusMethodTable server_methods[] = {
-	{ "SetProperty",	"sv",	"",	set_property },
-	{ "GetProperties",	"",	"a{sv}",get_properties },
+	{ "Register",	"ss",	"",	register_server		},
+	{ "Unregister",	"s",	"",	unregister_server	},
 	{ }
 };
 
-static GDBusSignalTable server_signals[] = {
-	{ "PropertyChanged",		"sv"		},
-	{ }
-};
-
-static struct network_adapter *create_adapter(const char *path, bdaddr_t *src)
+static struct network_adapter *create_adapter(struct btd_adapter *adapter)
 {
 	struct network_adapter *na;
-	int lm = 0;
-
-	if (security)
-		lm |= L2CAP_LM_AUTH | L2CAP_LM_ENCRYPT;
+	GError *err = NULL;
+	bdaddr_t src;
 
 	na = g_new0(struct network_adapter, 1);
-	na->path = g_strdup(path);
-	bacpy(&na->src, src);
+	na->adapter = btd_adapter_ref(adapter);
 
-	na->io = bt_l2cap_listen(src, BNEP_PSM, BNEP_MTU, lm,
-			connect_event, na);
+	adapter_get_address(adapter, &src);
+
+	na->io = bt_io_listen(BT_IO_L2CAP, NULL, confirm_event, na,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &src,
+				BT_IO_OPT_PSM, BNEP_PSM,
+				BT_IO_OPT_OMTU, BNEP_MTU,
+				BT_IO_OPT_IMTU, BNEP_MTU,
+				BT_IO_OPT_SEC_LEVEL,
+				security ? BT_IO_SEC_MEDIUM : BT_IO_SEC_LOW,
+				BT_IO_OPT_INVALID);
 	if (!na->io) {
+		error("%s", err->message);
+		g_error_free(err);
 		adapter_free(na);
 		return NULL;
 	}
 
-	g_io_channel_set_close_on_unref(na->io, FALSE);
-
 	return na;
 }
 
-int server_register(const char *path, bdaddr_t *src, uint16_t id)
+int server_register(struct btd_adapter *adapter)
 {
 	struct network_adapter *na;
 	struct network_server *ns;
+	const char *path;
 
-	na = find_adapter(adapters, path);
+	na = find_adapter(adapters, adapter);
 	if (!na) {
-		na = create_adapter(path, src);
+		na = create_adapter(adapter);
 		if (!na)
 			return -EINVAL;
 		adapters = g_slist_append(adapters, na);
 	}
 
-	ns = find_server(na->servers, id);
+	ns = find_server(na->servers, BNEP_SVC_NAP);
 	if (ns)
 		return 0;
 
 	ns = g_new0(struct network_server, 1);
 
-	switch (id) {
-	case BNEP_SVC_PANU:
-		ns->iface = g_strdup(NETWORK_PEER_INTERFACE);
-		ns->name = g_strdup("BlueZ PANU service");
-		break;
-	case BNEP_SVC_GN:
-		ns->iface = g_strdup(NETWORK_HUB_INTERFACE);
-		ns->name = g_strdup("BlueZ GN service");
-		break;
-	case BNEP_SVC_NAP:
-		ns->iface = g_strdup(NETWORK_ROUTER_INTERFACE);
-		ns->name = g_strdup("BlueZ NAP service");
-		break;
-	}
+	ns->iface = g_strdup(NETWORK_SERVER_INTERFACE);
+	ns->name = g_strdup("Network service");
+
+	path = adapter_get_path(adapter);
 
 	if (!g_dbus_register_interface(connection, path, ns->iface,
-					server_methods, server_signals, NULL,
+					server_methods, NULL, NULL,
 					ns, path_unregister)) {
 		error("D-Bus failed to register %s interface",
 				ns->iface);
@@ -923,23 +762,24 @@ int server_register(const char *path, bdaddr_t *src, uint16_t id)
 		return -1;
 	}
 
-	ns->id = id;
+	adapter_get_address(adapter, &ns->src);
+	ns->id = BNEP_SVC_NAP;
 	ns->na = na;
-	ns->record_id = register_server_record(ns);
-	ns->enable = TRUE;
+	ns->record_id = 0;
 	na->servers = g_slist_append(na->servers, ns);
 
-	info("Registered interface %s on path %s", ns->iface, path);
+	DBG("Registered interface %s on path %s", ns->iface, path);
 
 	return 0;
 }
 
-int server_unregister(const char *path, uint16_t id)
+int server_unregister(struct btd_adapter *adapter)
 {
 	struct network_adapter *na;
 	struct network_server *ns;
+	uint16_t id = BNEP_SVC_NAP;
 
-	na = find_adapter(adapters, path);
+	na = find_adapter(adapters, adapter);
 	if (!na)
 		return -EINVAL;
 
@@ -947,7 +787,8 @@ int server_unregister(const char *path, uint16_t id)
 	if (!ns)
 		return -EINVAL;
 
-	g_dbus_unregister_interface(connection, path, ns->iface);
+	g_dbus_unregister_interface(connection, adapter_get_path(adapter),
+					ns->iface);
 
 	return 0;
 }
