@@ -42,6 +42,7 @@
 
 #include "hcid.h"
 #include "sdpd.h"
+#include "btio.h"
 #include "adapter.h"
 #include "device.h"
 #include "plugin.h"
@@ -63,8 +64,27 @@ enum {
 	PENDING_NAME,
 };
 
+struct uuid_info {
+	uuid_t uuid;
+	uint8_t svc_hint;
+};
+
+struct bt_conn {
+	struct dev_info *dev;
+	bdaddr_t bdaddr;
+	uint16_t handle;
+	uint8_t loc_cap;
+	uint8_t loc_auth;
+	uint8_t rem_cap;
+	uint8_t rem_auth;
+	gboolean bonding_initiator;
+	gboolean secmode3;
+	GIOChannel *io; /* For raw L2CAP socket (bonding) */
+};
+
 static int max_dev = -1;
 static struct dev_info {
+	int id;
 	int sk;
 	bdaddr_t bdaddr;
 	char name[249];
@@ -80,6 +100,9 @@ static struct dev_info {
 	gboolean cache_enable;
 	gboolean already_up;
 	gboolean registered;
+	gboolean pairable;
+
+	uint8_t io_capability;
 
 	struct hci_version ver;
 
@@ -95,9 +118,11 @@ static struct dev_info {
 
 	gboolean debug_keys;
 	GSList *keys;
-	int pin_length;
+	uint8_t pin_length;
 
 	GSList *uuids;
+
+	GSList *connections;
 } *devs = NULL;
 
 static int ignore_device(struct hci_dev_info *di)
@@ -105,16 +130,21 @@ static int ignore_device(struct hci_dev_info *di)
 	return hci_test_bit(HCI_RAW, &di->flags) || di->type >> 4 != HCI_BREDR;
 }
 
-static void init_dev_info(int index, int sk, gboolean registered)
+static struct dev_info *init_dev_info(int index, int sk, gboolean registered,
+							gboolean already_up)
 {
 	struct dev_info *dev = &devs[index];
 
 	memset(dev, 0, sizeof(*dev));
 
+	dev->id = index;
 	dev->sk = sk;
-	dev->pin_length = -1;
 	dev->cache_enable = TRUE;
 	dev->registered = registered;
+	dev->already_up = already_up;
+	dev->io_capability = 0x03; /* No Input No Output */
+
+	return dev;
 }
 
 /* Async HCI command handling with callback support */
@@ -318,20 +348,17 @@ static int hciops_set_pairable(int index, gboolean pairable)
 	if (adapter)
 		btd_adapter_pairable_changed(adapter, pairable);
 
+	devs[index].pairable = pairable;
+
 	return 0;
 }
 
 static int hciops_power_off(int index)
 {
 	struct dev_info *dev = &devs[index];
-	uint8_t mode = SCAN_DISABLED;
 
 	DBG("hci%d", index);
 
-	if (hci_send_cmd(dev->sk, OGF_HOST_CTL,
-					OCF_WRITE_SCAN_ENABLE, 1, &mode) < 0)
-		return -errno;
-
 	if (ioctl(dev->sk, HCIDEVDOWN, index) < 0 && errno != EALREADY)
 		return -errno;
 
@@ -432,6 +459,29 @@ static void start_adapter(int index)
 	memset(dev->eir, 0, sizeof(dev->eir));
 }
 
+static int hciops_stop_inquiry(int index)
+{
+	struct dev_info *dev = &devs[index];
+	struct hci_dev_info di;
+	int err;
+
+	DBG("hci%d", index);
+
+	if (hci_devinfo(index, &di) < 0)
+		return -errno;
+
+	if (hci_test_bit(HCI_INQUIRY, &di.flags))
+		err = hci_send_cmd(dev->sk, OGF_LINK_CTL,
+						OCF_INQUIRY_CANCEL, 0, 0);
+	else
+		err = hci_send_cmd(dev->sk, OGF_LINK_CTL,
+					OCF_EXIT_PERIODIC_INQUIRY, 0, 0);
+	if (err < 0)
+		err = -errno;
+
+	return err;
+}
+
 static gboolean init_adapter(int index)
 {
 	struct dev_info *dev = &devs[index];
@@ -471,6 +521,9 @@ static gboolean init_adapter(int index)
 	hciops_set_discoverable(index, discoverable);
 	hciops_set_pairable(index, pairable);
 
+	if (dev->already_up)
+		hciops_stop_inquiry(index);
+
 done:
 	btd_adapter_unref(adapter);
 	return TRUE;
@@ -568,31 +621,83 @@ static int hciops_set_did(int index, uint16_t vendor, uint16_t product,
 
 /* Start of HCI event callbacks */
 
-static int get_handle(int index, const bdaddr_t *bdaddr, uint16_t *handle)
+static gint conn_handle_cmp(gconstpointer a, gconstpointer b)
+{
+	const struct bt_conn *conn = a;
+	uint16_t handle = *((const uint16_t *) b);
+
+	return (int) conn->handle - (int) handle;
+}
+
+static struct bt_conn *find_conn_by_handle(struct dev_info *dev,
+							uint16_t handle)
+{
+	GSList *match;
+
+	match = g_slist_find_custom(dev->connections, &handle,
+							conn_handle_cmp);
+	if (match)
+		return match->data;
+
+	return NULL;
+}
+
+static gint conn_bdaddr_cmp(gconstpointer a, gconstpointer b)
+{
+	const struct bt_conn *conn = a;
+	const bdaddr_t *bdaddr = b;
+
+	return bacmp(&conn->bdaddr, bdaddr);
+}
+
+static struct bt_conn *find_connection(struct dev_info *dev, bdaddr_t *bdaddr)
+{
+	GSList *match;
+
+	match = g_slist_find_custom(dev->connections, bdaddr, conn_bdaddr_cmp);
+	if (match)
+		return match->data;
+
+	return NULL;
+}
+
+static struct bt_conn *get_connection(struct dev_info *dev, bdaddr_t *bdaddr)
+{
+	struct bt_conn *conn;
+
+	conn = find_connection(dev, bdaddr);
+	if (conn)
+		return conn;
+
+	conn = g_new0(struct bt_conn, 1);
+
+	conn->dev = dev;
+	conn->loc_cap = dev->io_capability;
+	conn->loc_auth = 0xff;
+	conn->rem_auth = 0xff;
+	bacpy(&conn->bdaddr, bdaddr);
+
+	dev->connections = g_slist_append(dev->connections, conn);
+
+	return conn;
+}
+
+static int get_handle(int index, bdaddr_t *bdaddr, uint16_t *handle)
 {
 	struct dev_info *dev = &devs[index];
-	struct hci_conn_info_req *cr;
+	struct bt_conn *conn;
 	char addr[18];
-	int err;
 
 	ba2str(bdaddr, addr);
 	DBG("hci%d dba %s", index, addr);
 
-	cr = g_malloc0(sizeof(*cr) + sizeof(struct hci_conn_info));
-	bacpy(&cr->bdaddr, bdaddr);
-	cr->type = ACL_LINK;
+	conn = find_connection(dev, bdaddr);
+	if (conn == NULL)
+		return -ENOENT;
 
-	if (ioctl(dev->sk, HCIGETCONNINFO, (unsigned long) cr) < 0) {
-		err = -errno;
-		goto fail;
-	}
+	*handle = conn->handle;
 
-	err = 0;
-	*handle = cr->conn_info->handle;
-
-fail:
-	g_free(cr);
-	return err;
+	return 0;
 }
 
 static int disconnect_addr(int index, bdaddr_t *dba, uint8_t reason)
@@ -616,34 +721,43 @@ static int disconnect_addr(int index, bdaddr_t *dba, uint8_t reason)
 	return 0;
 }
 
-static inline int get_bdaddr(int index, uint16_t handle, bdaddr_t *dba)
+static void bonding_complete(struct dev_info *dev, struct bt_conn *conn,
+								uint8_t status)
 {
-	struct dev_info *dev = &devs[index];
-	struct hci_conn_list_req *cl;
-	struct hci_conn_info *ci;
-	int i;
+	DBG("status 0x%02x", status);
 
-	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
-
-	cl->dev_id = index;
-	cl->conn_num = 10;
-	ci = cl->conn_info;
-
-	if (ioctl(dev->sk, HCIGETCONNLIST, (void *) cl) < 0) {
-		g_free(cl);
-		return -EIO;
+	if (conn->io != NULL) {
+		/* bonding_connect_cb takes care of the successul case */
+		if (status != 0)
+			g_io_channel_shutdown(conn->io, TRUE, NULL);
+		g_io_channel_unref(conn->io);
+		conn->io = NULL;
 	}
 
-	for (i = 0; i < cl->conn_num; i++, ci++)
-		if (ci->handle == handle) {
-			bacpy(dba, &ci->bdaddr);
-			g_free(cl);
-			return 0;
-		}
+	conn->bonding_initiator = FALSE;
 
-	g_free(cl);
+	btd_event_bonding_complete(&dev->bdaddr, &conn->bdaddr, status);
+}
 
-	return -ENOENT;
+static int get_auth_info(int index, bdaddr_t *bdaddr, uint8_t *auth)
+{
+	struct dev_info *dev = &devs[index];
+	struct hci_auth_info_req req;
+	char addr[18];
+
+	ba2str(bdaddr, addr);
+	DBG("hci%d dba %s", index, addr);
+
+	memset(&req, 0, sizeof(req));
+	bacpy(&req.bdaddr, bdaddr);
+
+	if (ioctl(dev->sk, HCIGETAUTHINFO, (unsigned long) &req) < 0)
+		return -errno;
+
+	if (auth)
+		*auth = req.type;
+
+	return 0;
 }
 
 /* Link Key handling */
@@ -651,27 +765,21 @@ static inline int get_bdaddr(int index, uint16_t handle, bdaddr_t *dba)
 static void link_key_request(int index, bdaddr_t *dba)
 {
 	struct dev_info *dev = &devs[index];
-	struct hci_auth_info_req req;
-	GSList *match;
 	struct link_key_info *key_info;
+	struct bt_conn *conn;
+	GSList *match;
 	char da[18];
-	int err;
 
 	ba2str(dba, da);
 	DBG("hci%d dba %s", index, da);
 
-	memset(&req, 0, sizeof(req));
-	bacpy(&req.bdaddr, dba);
+	conn = get_connection(dev, dba);
+	if (conn->handle == 0)
+		conn->secmode3 = TRUE;
 
-	err = ioctl(dev->sk, HCIGETAUTHINFO, (unsigned long) &req);
-	if (err < 0) {
-		if (errno != EINVAL)
-			DBG("HCIGETAUTHINFO failed %s (%d)",
-						strerror(errno), errno);
-		req.type = 0x00;
-	}
+	get_auth_info(index, dba, &conn->loc_auth);
 
-	DBG("kernel auth requirements = 0x%02x", req.type);
+	DBG("kernel auth requirements = 0x%02x", conn->loc_auth);
 
 	match = g_slist_find_custom(dev->keys, dba, (GCompareFunc) bacmp);
 	if (match)
@@ -694,7 +802,8 @@ static void link_key_request(int index, bdaddr_t *dba)
 
 	/* Don't use unauthenticated combination keys if MITM is
 	 * required */
-	if (key_info->type == 0x04 && req.type != 0xff && (req.type & 0x01))
+	if (key_info->type == 0x04 && conn->loc_auth != 0xff &&
+						(conn->loc_auth & 0x01))
 		hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_LINK_KEY_NEG_REPLY,
 								6, dba);
 	else {
@@ -714,14 +823,17 @@ static void link_key_notify(int index, void *ptr)
 	evt_link_key_notify *evt = ptr;
 	bdaddr_t *dba = &evt->bdaddr;
 	struct link_key_info *key_info;
+	uint8_t old_key_type, key_type;
+	struct bt_conn *conn;
 	GSList *match;
-	uint8_t old_key_type, reason;
 	char da[18];
-	int err;
+	uint8_t status = 0;
 
 	ba2str(dba, da);
 	DBG("hci%d dba %s type %d", index, da, evt->key_type);
 
+	conn = get_connection(dev, &evt->bdaddr);
+
 	match = g_slist_find_custom(dev->keys, dba, (GCompareFunc) bacmp);
 	if (match)
 		key_info = match->data;
@@ -741,26 +853,77 @@ static void link_key_notify(int index, void *ptr)
 	key_info->type = evt->key_type;
 	key_info->pin_len = dev->pin_length;
 
-	err = btd_event_link_key_notify(&dev->bdaddr, dba, evt->link_key,
-					evt->key_type, dev->pin_length,
-					old_key_type);
-	dev->pin_length = -1;
+	key_type = evt->key_type;
 
-	if (err == 0) {
-		dev->keys = g_slist_append(dev->keys, key_info);
+	DBG("key type 0x%02x old key type 0x%02x", key_type, old_key_type);
+	DBG("local auth 0x%02x and remote auth 0x%02x",
+					conn->loc_auth, conn->rem_auth);
+
+	if (key_type == 0x06) {
+		/* Some buggy controller combinations generate a changed
+		 * combination key for legacy pairing even when there's no
+		 * previous key */
+		if ((!conn || conn->rem_auth == 0xff) && old_key_type == 0xff)
+			key_type = 0x00;
+		else if (old_key_type != 0xff)
+			key_type = old_key_type;
+		else
+			/* This is Changed Combination Link Key for
+			 * a temporary link key.*/
+			goto done;
+	}
+
+	key_info->type = key_type;
+
+	/* Skip the storage check if this is a debug key */
+	if (key_type == 0x03)
+		goto done;
+
+	/* Store the link key persistently if one of the following is true:
+	 * 1. this is a legacy link key
+	 * 2. this is a changed combination key and there was a previously
+	 *    stored one
+	 * 3. neither local nor remote side had no-bonding as a requirement
+	 * 4. the local side had dedicated bonding as a requirement
+	 * 5. the remote side is using dedicated bonding since in that case
+	 *    also the local requirements are set to dedicated bonding
+	 * If none of the above match only keep the link key around for
+	 * this connection and set the temporary flag for the device.
+	 */
+	if (key_type < 0x03 || (key_type == 0x06 && old_key_type != 0xff) ||
+			(conn->loc_auth > 0x01 && conn->rem_auth > 0x01) ||
+			(conn->loc_auth == 0x02 || conn->loc_auth == 0x03) ||
+			(conn->rem_auth == 0x02 || conn->rem_auth == 0x03)) {
+		int err;
+
+		err = btd_event_link_key_notify(&dev->bdaddr, dba,
+						evt->link_key, key_type,
+						dev->pin_length);
+
+		if (err == -ENODEV)
+			status = HCI_OE_LOW_RESOURCES;
+		else if (err < 0)
+			status = HCI_MEMORY_FULL;
+
+		goto done;
+	}
+
+done:
+	dev->pin_length = 0;
+
+	if (status != 0) {
+		g_free(key_info);
+		bonding_complete(dev, conn, status);
+		disconnect_addr(index, dba, status);
 		return;
 	}
 
-	g_free(key_info);
+	dev->keys = g_slist_prepend(dev->keys, key_info);
 
-	if (err == -ENODEV)
-		reason = HCI_OE_LOW_RESOURCES;
-	else
-		reason = HCI_MEMORY_FULL;
-
-	btd_event_bonding_process_complete(&dev->bdaddr, dba, reason);
-
-	disconnect_addr(index, dba, reason);
+	/* If we're connected and not dedicated bonding initiators we're
+	 * done with the bonding process */
+	if (!conn->bonding_initiator && conn->handle != 0)
+		bonding_complete(dev, conn, 0);
 }
 
 static void return_link_keys(int index, void *ptr)
@@ -795,13 +958,42 @@ static void user_confirm_request(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_user_confirm_request *req = ptr;
+	gboolean loc_mitm, rem_mitm, auto_accept;
+	struct bt_conn *conn;
 
 	DBG("hci%d", index);
 
+	conn = find_connection(dev, &req->bdaddr);
+	if (conn == NULL)
+		return;
+
+	loc_mitm = (conn->loc_auth & 0x01) ? TRUE : FALSE;
+	rem_mitm = (conn->rem_auth & 0x01) ? TRUE : FALSE;
+
+	/* If we require MITM but the remote device can't provide that
+	 * (it has NoInputNoOutput) then reject the confirmation
+	 * request. The only exception is when we're dedicated bonding
+	 * initiators since then we always have the MITM bit set. */
+	if (!conn->bonding_initiator && loc_mitm && conn->rem_cap == 0x03) {
+		error("Rejecting request: remote device can't provide MITM");
+		goto fail;
+	}
+
+	/* If no side requires MITM protection; auto-accept */
+	if ((conn->loc_auth == 0xff || !loc_mitm || conn->rem_cap == 0x03) &&
+					(!rem_mitm || conn->loc_cap == 0x03)) {
+		DBG("auto accept of confirmation");
+		auto_accept = TRUE;
+	} else
+		auto_accept = FALSE;
+
 	if (btd_event_user_confirm(&dev->bdaddr, &req->bdaddr,
-					btohl(req->passkey)) < 0)
-		hci_send_cmd(dev->sk, OGF_LINK_CTL,
-				OCF_USER_CONFIRM_NEG_REPLY, 6, ptr);
+				btohl(req->passkey), auto_accept) == 0)
+		return;
+
+fail:
+	hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_USER_CONFIRM_NEG_REPLY,
+								6, ptr);
 }
 
 static void user_passkey_request(int index, void *ptr)
@@ -837,18 +1029,93 @@ static void remote_oob_data_request(int index, void *ptr)
 				OCF_REMOTE_OOB_DATA_NEG_REPLY, 6, ptr);
 }
 
+static int get_io_cap(int index, bdaddr_t *bdaddr, uint8_t *cap, uint8_t *auth)
+{
+	struct dev_info *dev = &devs[index];
+	struct bt_conn *conn;
+	int err;
+
+	conn = find_connection(dev, bdaddr);
+	if (conn == NULL)
+		return -ENOENT;
+
+	err = get_auth_info(index, bdaddr, &conn->loc_auth);
+	if (err < 0)
+		return err;
+
+	DBG("initial authentication requirement is 0x%02x", conn->loc_auth);
+
+	if (!dev->pairable && !conn->bonding_initiator) {
+		if (conn->rem_auth < 0x02) {
+			DBG("Allowing no bonding in non-bondable mode");
+			/* Kernel defaults to general bonding and so
+			 * overwrite for this special case. Otherwise
+			 * non-pairable test cases will fail. */
+			conn->loc_auth = conn->rem_auth;
+			goto done;
+		}
+
+		return -EPERM;
+	}
+
+	/* If the kernel doesn't know the local requirement just mirror
+	 * the remote one */
+	if (conn->loc_auth == 0xff)
+		conn->loc_auth = conn->rem_auth;
+
+	if (conn->loc_auth == 0x00 || conn->loc_auth == 0x04) {
+		/* If remote requests dedicated bonding follow that lead */
+		if (conn->rem_auth == 0x02 || conn->rem_auth == 0x03) {
+
+			/* If both remote and local IO capabilities allow MITM
+			 * then require it, otherwise don't */
+			if (conn->rem_cap == 0x03 || conn->loc_cap == 0x03)
+				conn->loc_auth = 0x02;
+			else
+				conn->loc_auth = 0x03;
+		}
+
+		/* If remote indicates no bonding then follow that. This
+		 * is important since the kernel might give general bonding
+		 * as default. */
+		if (conn->rem_auth == 0x00 || conn->rem_auth == 0x01)
+			conn->loc_auth = 0x00;
+
+		/* If remote requires MITM then also require it, unless
+		 * our IO capability is NoInputNoOutput (so some
+		 * just-works security cases can be tested) */
+		if (conn->rem_auth != 0xff && (conn->rem_auth & 0x01) &&
+							conn->loc_cap != 0x03)
+			conn->loc_auth |= 0x01;
+	}
+
+done:
+	*cap = conn->loc_cap;
+	*auth = conn->loc_auth;
+
+	DBG("final authentication requirement is 0x%02x", *auth);
+
+	return 0;
+}
+
 static void io_capa_request(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	bdaddr_t *dba = ptr;
+	uint8_t cap, auth = 0xff;
 	char da[18];
-	uint8_t cap, auth;
+	int err;
 
 	ba2str(dba, da);
 	DBG("hci%d IO capability request for %s", index, da);
 
-	if (btd_event_get_io_cap(&dev->bdaddr, dba, &cap, &auth) < 0) {
+	err = get_io_cap(index, dba, &cap, &auth);
+	if (err < 0) {
 		io_capability_neg_reply_cp cp;
+
+		error("Getting IO capability failed: %s (%d)",
+						strerror(-err), -err);
+
 		memset(&cp, 0, sizeof(cp));
 		bacpy(&cp.bdaddr, dba);
 		cp.reason = HCI_PAIRING_NOT_ALLOWED;
@@ -871,13 +1138,17 @@ static void io_capa_response(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_io_capability_response *evt = ptr;
+	struct bt_conn *conn;
 	char da[18];
 
 	ba2str(&evt->bdaddr, da);
 	DBG("hci%d IO capability response from %s", index, da);
 
-	btd_event_set_io_cap(&dev->bdaddr, &evt->bdaddr,
-				evt->capability, evt->authentication);
+	conn = find_connection(dev, &evt->bdaddr);
+	if (conn) {
+		conn->rem_cap = evt->capability;
+		conn->rem_auth = evt->authentication;
+	}
 }
 
 /* PIN code handling */
@@ -885,18 +1156,34 @@ static void io_capa_response(int index, void *ptr)
 static void pin_code_request(int index, bdaddr_t *dba)
 {
 	struct dev_info *dev = &devs[index];
+	struct bt_conn *conn;
 	char addr[18];
 	int err;
 
 	ba2str(dba, addr);
 	DBG("hci%d PIN request for %s", index, addr);
 
+	conn = get_connection(dev, dba);
+	if (conn->handle == 0)
+		conn->secmode3 = TRUE;
+
+	/* Check if the adapter is not pairable and if there isn't a bonding in
+	 * progress */
+	if (!dev->pairable && !conn->bonding_initiator) {
+		DBG("Rejecting PIN request in non-pairable mode");
+		goto reject;
+	}
+
 	err = btd_event_request_pin(&dev->bdaddr, dba);
 	if (err < 0) {
 		error("PIN code negative reply: %s", strerror(-err));
-		hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY,
-								6, dba);
+		goto reject;
 	}
+
+	return;
+
+reject:
+	hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, dba);
 }
 
 static void start_inquiry(bdaddr_t *local, uint8_t status, gboolean periodic)
@@ -1037,10 +1324,10 @@ static void eir_generate_uuid128(GSList *list, uint8_t *ptr, uint16_t *eir_len)
 	uuid128 = ptr + 2;
 
 	for (; list; list = list->next) {
-		uuid_t *uuid = list->data;
-		uint8_t *uuid128_data = uuid->value.uuid128.data;
+		struct uuid_info *uuid = list->data;
+		uint8_t *uuid128_data = uuid->uuid.value.uuid128.data;
 
-		if (uuid->type != SDP_UUID128)
+		if (uuid->uuid.type != SDP_UUID128)
 			continue;
 
 		/* Stop if not enough space to put next UUID128 */
@@ -1136,15 +1423,15 @@ static void create_ext_inquiry_response(int index, uint8_t *data)
 
 	/* Group all UUID16 types */
 	for (l = dev->uuids; l != NULL; l = g_slist_next(l)) {
-		uuid_t *uuid = l->data;
+		struct uuid_info *uuid = l->data;
 
-		if (uuid->type != SDP_UUID16)
+		if (uuid->uuid.type != SDP_UUID16)
 			continue;
 
-		if (uuid->value.uuid16 < 0x1100)
+		if (uuid->uuid.value.uuid16 < 0x1100)
 			continue;
 
-		if (uuid->value.uuid16 == PNP_INFO_SVCLASS_ID)
+		if (uuid->uuid.value.uuid16 == PNP_INFO_SVCLASS_ID)
 			continue;
 
 		/* Stop if not enough space to put next UUID16 */
@@ -1155,13 +1442,13 @@ static void create_ext_inquiry_response(int index, uint8_t *data)
 
 		/* Check for duplicates */
 		for (i = 0; i < uuid_count; i++)
-			if (uuid16[i] == uuid->value.uuid16)
+			if (uuid16[i] == uuid->uuid.value.uuid16)
 				break;
 
 		if (i < uuid_count)
 			continue;
 
-		uuid16[uuid_count++] = uuid->value.uuid16;
+		uuid16[uuid_count++] = uuid->uuid.value.uuid16;
 		eir_len += sizeof(uint16_t);
 	}
 
@@ -1562,18 +1849,20 @@ static inline void remote_version_information(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_read_remote_version_complete *evt = ptr;
-	bdaddr_t dba;
+	struct bt_conn *conn;
 
 	DBG("hci%d status %u", index, evt->status);
 
 	if (evt->status)
 		return;
 
-	if (get_bdaddr(index, btohs(evt->handle), &dba) < 0)
+	conn = find_conn_by_handle(dev, btohs(evt->handle));
+	if (conn == NULL)
 		return;
 
-	write_version_info(&dev->bdaddr, &dba, btohs(evt->manufacturer),
-				evt->lmp_ver, btohs(evt->lmp_subver));
+	write_version_info(&dev->bdaddr, &conn->bdaddr,
+				btohs(evt->manufacturer), evt->lmp_ver,
+				btohs(evt->lmp_subver));
 }
 
 static inline void inquiry_result(int index, int plen, void *ptr)
@@ -1650,17 +1939,18 @@ static inline void remote_features_information(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_read_remote_features_complete *evt = ptr;
-	bdaddr_t dba;
+	struct bt_conn *conn;
 
 	DBG("hci%d status %u", index, evt->status);
 
 	if (evt->status)
 		return;
 
-	if (get_bdaddr(index, btohs(evt->handle), &dba) < 0)
+	conn = find_conn_by_handle(dev, btohs(evt->handle));
+	if (conn == NULL)
 		return;
 
-	write_features_info(&dev->bdaddr, &dba, evt->features, NULL);
+	write_features_info(&dev->bdaddr, &conn->bdaddr, evt->features, NULL);
 }
 
 struct remote_version_req {
@@ -1697,21 +1987,58 @@ static void get_remote_version(int index, uint16_t handle)
 								req, g_free);
 }
 
+static void conn_free(struct bt_conn *conn)
+{
+	if (conn->io != NULL) {
+		g_io_channel_shutdown(conn->io, TRUE, NULL);
+		g_io_channel_unref(conn->io);
+	}
+
+	g_free(conn);
+}
+
+static inline void conn_failed(int index, bdaddr_t *bdaddr, uint8_t status)
+{
+	struct dev_info *dev = &devs[index];
+	struct bt_conn *conn;
+
+	btd_event_conn_failed(&dev->bdaddr, bdaddr, status);
+
+	conn = find_connection(dev, bdaddr);
+	if (conn == NULL)
+		return;
+
+	bonding_complete(dev, conn, status);
+
+	dev->connections = g_slist_remove(dev->connections, conn);
+	conn_free(conn);
+}
+
 static inline void conn_complete(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_conn_complete *evt = ptr;
 	char filename[PATH_MAX];
 	char local_addr[18], peer_addr[18], *str;
+	struct bt_conn *conn;
 
 	if (evt->link_type != ACL_LINK)
 		return;
 
-	btd_event_conn_complete(&dev->bdaddr, evt->status,
-					btohs(evt->handle), &evt->bdaddr);
+	DBG("status 0x%02x", evt->status);
 
-	if (evt->status)
+	if (evt->status != 0) {
+		conn_failed(index, &evt->bdaddr, evt->status);
 		return;
+	}
+
+	conn = get_connection(dev, &evt->bdaddr);
+	conn->handle = btohs(evt->handle);
+
+	btd_event_conn_complete(&dev->bdaddr, &evt->bdaddr);
+
+	if (conn->secmode3)
+		bonding_complete(dev, conn, 0);
 
 	/* check if the remote version needs be requested */
 	ba2str(&dev->bdaddr, local_addr);
@@ -1733,12 +2060,18 @@ static inline void le_conn_complete(int index, void *ptr)
 	evt_le_connection_complete *evt = ptr;
 	char filename[PATH_MAX];
 	char local_addr[18], peer_addr[18], *str;
+	struct bt_conn *conn;
 
-	btd_event_conn_complete(&dev->bdaddr, evt->status,
-					btohs(evt->handle), &evt->peer_bdaddr);
-
-	if (evt->status)
+	if (evt->status) {
+		btd_event_conn_failed(&dev->bdaddr, &evt->peer_bdaddr,
+								evt->status);
 		return;
+	}
+
+	conn = get_connection(dev, &evt->peer_bdaddr);
+	conn->handle = btohs(evt->handle);
+
+	btd_event_conn_complete(&dev->bdaddr, &evt->peer_bdaddr);
 
 	/* check if the remote version needs be requested */
 	ba2str(&dev->bdaddr, local_addr);
@@ -1758,23 +2091,37 @@ static inline void disconn_complete(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_disconn_complete *evt = ptr;
+	struct bt_conn *conn;
 
-	btd_event_disconn_complete(&dev->bdaddr, evt->status,
-					btohs(evt->handle), evt->reason);
+	DBG("handle %u status 0x%02x", btohs(evt->handle), evt->status);
+
+	if (evt->status != 0)
+		return;
+
+	conn = find_conn_by_handle(dev, btohs(evt->handle));
+	if (conn == NULL)
+		return;
+
+	dev->connections = g_slist_remove(dev->connections, conn);
+
+	btd_event_disconn_complete(&dev->bdaddr, &conn->bdaddr);
+
+	conn_free(conn);
 }
 
 static inline void auth_complete(int index, void *ptr)
 {
 	struct dev_info *dev = &devs[index];
 	evt_auth_complete *evt = ptr;
-	bdaddr_t dba;
+	struct bt_conn *conn;
 
 	DBG("hci%d status %u", index, evt->status);
 
-	if (get_bdaddr(index, btohs(evt->handle), &dba) < 0)
+	conn = find_conn_by_handle(dev, btohs(evt->handle));
+	if (conn == NULL)
 		return;
 
-	btd_event_bonding_process_complete(&dev->bdaddr, &dba, evt->status);
+	bonding_complete(dev, conn, evt->status);
 }
 
 static inline void simple_pairing_complete(int index, void *ptr)
@@ -1853,7 +2200,10 @@ static void stop_hci_dev(int index)
 	g_slist_foreach(dev->uuids, (GFunc) g_free, NULL);
 	g_slist_free(dev->uuids);
 
-	init_dev_info(index, -1, dev->registered);
+	g_slist_foreach(dev->connections, (GFunc) conn_free, NULL);
+	g_slist_free(dev->connections);
+
+	init_dev_info(index, -1, dev->registered, dev->already_up);
 }
 
 static gboolean io_security_event(GIOChannel *chan, GIOCondition cond,
@@ -1863,18 +2213,21 @@ static gboolean io_security_event(GIOChannel *chan, GIOCondition cond,
 	int type, index = GPOINTER_TO_INT(data);
 	struct dev_info *dev = &devs[index];
 	struct hci_dev_info di;
-	size_t len;
+	ssize_t len;
 	hci_event_hdr *eh;
-	GIOError err;
 	evt_cmd_status *evt;
+	int fd;
 
 	if (cond & (G_IO_NVAL | G_IO_HUP | G_IO_ERR)) {
 		stop_hci_dev(index);
 		return FALSE;
 	}
 
-	if ((err = g_io_channel_read(chan, (gchar *) buf, sizeof(buf), &len))) {
-		if (err == G_IO_ERROR_AGAIN)
+	fd = g_io_channel_unix_get_fd(chan);
+
+	len = read(fd, buf, sizeof(buf));
+	if (len < 0) {
+		if (errno == EAGAIN)
 			return TRUE;
 		stop_hci_dev(index);
 		return FALSE;
@@ -2058,7 +2411,7 @@ static void start_hci_dev(int index)
 						io_security_event,
 						GINT_TO_POINTER(index), NULL);
 	dev->io = chan;
-	dev->pin_length = -1;
+	dev->pin_length = 0;
 
 }
 
@@ -2135,8 +2488,9 @@ static void init_pending(int index)
 	hci_set_bit(PENDING_NAME, &dev->pending);
 }
 
-static void init_device(int index)
+static struct dev_info *init_device(int index, gboolean already_up)
 {
+	struct dev_info *dev;
 	struct hci_dev_req dr;
 	int dd;
 	pid_t pid;
@@ -2147,7 +2501,7 @@ static void init_device(int index)
 	if (dd < 0) {
 		error("Unable to open hci%d: %s (%d)", index,
 						strerror(errno), errno);
-		return;
+		return NULL;
 	}
 
 	if (index > max_dev) {
@@ -2155,10 +2509,14 @@ static void init_device(int index)
 		devs = g_realloc(devs, sizeof(devs[0]) * (max_dev + 1));
 	}
 
-	init_dev_info(index, dd, FALSE);
+	dev = init_dev_info(index, dd, FALSE, already_up);
 	init_pending(index);
 	start_hci_dev(index);
 
+	/* Avoid forking if nothing else has to be done */
+	if (already_up)
+		return dev;
+
 	/* Do initialization in the separate process */
 	pid = fork();
 	switch (pid) {
@@ -2170,7 +2528,7 @@ static void init_device(int index)
 					index, strerror(errno), errno);
 		default:
 			DBG("child %d forked", pid);
-			return;
+			return dev;
 	}
 
 	memset(&dr, 0, sizeof(dr));
@@ -2197,23 +2555,41 @@ fail:
 	exit(1);
 }
 
-static void device_devreg_setup(int index)
+static void init_conn_list(int index)
 {
-	struct hci_dev_info di;
+	struct dev_info *dev = &devs[index];
+	struct hci_conn_list_req *cl;
+	struct hci_conn_info *ci;
+	int err, i;
 
 	DBG("hci%d", index);
 
-	init_device(index);
+	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
 
-	memset(&di, 0, sizeof(di));
+	cl->dev_id = index;
+	cl->conn_num = 10;
+	ci = cl->conn_info;
 
-	if (hci_devinfo(index, &di) < 0)
-		return;
+	if (ioctl(dev->sk, HCIGETCONNLIST, cl) < 0) {
+		error("Unable to get connection list: %s (%d)",
+						strerror(errno), errno);
+		goto failed;
+	}
 
-	if (ignore_device(&di))
-		return;
+	for (i = 0; i < cl->conn_num; i++, ci++) {
+		struct bt_conn *conn;
 
-	devs[index].already_up = hci_test_bit(HCI_UP, &di.flags);
+		if (ci->type != ACL_LINK)
+			continue;
+
+		conn = get_connection(dev, &ci->bdaddr);
+		conn->handle = ci->handle;
+	}
+
+	err = 0;
+
+failed:
+	g_free(cl);
 }
 
 static void device_event(int event, int index)
@@ -2221,9 +2597,7 @@ static void device_event(int event, int index)
 	switch (event) {
 	case HCI_DEV_REG:
 		info("HCI dev %d registered", index);
-		device_devreg_setup(index);
-		if (devs[index].already_up)
-			device_event(HCI_DEV_UP, index);
+		init_device(index, FALSE);
 		break;
 
 	case HCI_DEV_UNREG:
@@ -2257,29 +2631,6 @@ static void device_event(int event, int index)
 	}
 }
 
-static int hciops_stop_inquiry(int index)
-{
-	struct dev_info *dev = &devs[index];
-	struct hci_dev_info di;
-	int err;
-
-	DBG("hci%d", index);
-
-	if (hci_devinfo(index, &di) < 0)
-		return -errno;
-
-	if (hci_test_bit(HCI_INQUIRY, &di.flags))
-		err = hci_send_cmd(dev->sk, OGF_LINK_CTL,
-						OCF_INQUIRY_CANCEL, 0, 0);
-	else
-		err = hci_send_cmd(dev->sk, OGF_LINK_CTL,
-					OCF_EXIT_PERIODIC_INQUIRY, 0, 0);
-	if (err < 0)
-		err = -errno;
-
-	return err;
-}
-
 static gboolean init_known_adapters(gpointer user_data)
 {
 	struct hci_dev_list_req *dl;
@@ -2309,17 +2660,18 @@ static gboolean init_known_adapters(gpointer user_data)
 
 	for (i = 0; i < dl->dev_num; i++, dr++) {
 		struct dev_info *dev;
+		gboolean already_up;
 
-		device_event(HCI_DEV_REG, dr->dev_id);
+		already_up = hci_test_bit(HCI_UP, &dr->dev_opt);
 
-		dev = &devs[dr->dev_id];
-
-		dev->already_up = hci_test_bit(HCI_UP, &dr->dev_opt);
+		dev = init_device(dr->dev_id, already_up);
+		if (dev == NULL)
+			continue;
 
 		if (!dev->already_up)
 			continue;
 
-		hciops_stop_inquiry(dr->dev_id);
+		init_conn_list(dr->dev_id);
 
 		dev->pending = 0;
 		hci_set_bit(PENDING_VERSION, &dev->pending);
@@ -2340,15 +2692,16 @@ static gboolean io_stack_event(GIOChannel *chan, GIOCondition cond,
 	evt_stack_internal *si;
 	evt_si_device *sd;
 	hci_event_hdr *eh;
-	int type;
-	size_t len;
-	GIOError err;
+	int type, fd;
+	ssize_t len;
 
 	ptr = buf;
 
-	err = g_io_channel_read(chan, (gchar *) buf, sizeof(buf), &len);
-	if (err) {
-		if (err == G_IO_ERROR_AGAIN)
+	fd = g_io_channel_unix_get_fd(chan);
+
+	len = read(fd, buf, sizeof(buf));
+	if (len < 0) {
+		if (errno == EAGAIN)
 			return TRUE;
 
 		error("Read from control socket failed: %s (%d)",
@@ -2765,32 +3118,20 @@ static int hciops_unblock_device(int index, bdaddr_t *bdaddr)
 static int hciops_get_conn_list(int index, GSList **conns)
 {
 	struct dev_info *dev = &devs[index];
-	struct hci_conn_list_req *cl;
-	struct hci_conn_info *ci;
-	int err, i;
+	GSList *l;
 
 	DBG("hci%d", index);
 
-	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
-
-	cl->dev_id = index;
-	cl->conn_num = 10;
-	ci = cl->conn_info;
-
-	if (ioctl(dev->sk, HCIGETCONNLIST, cl) < 0) {
-		err = -errno;
-		goto fail;
-	}
-
-	err = 0;
 	*conns = NULL;
 
-	for (i = 0; i < cl->conn_num; i++, ci++)
-		*conns = g_slist_append(*conns, g_memdup(ci, sizeof(*ci)));
+	for (l = dev->connections; l != NULL; l = g_slist_next(l)) {
+		struct bt_conn *conn = l->data;
 
-fail:
-	g_free(cl);
-	return err;
+		*conns = g_slist_append(*conns,
+				g_memdup(&conn->bdaddr, sizeof(bdaddr_t)));
+	}
+
+	return 0;
 }
 
 static int hciops_read_local_version(int index, struct hci_version *ver)
@@ -2815,22 +3156,11 @@ static int hciops_read_local_features(int index, uint8_t *features)
 	return  0;
 }
 
-static int hciops_disconnect(int index, uint16_t handle)
+static int hciops_disconnect(int index, bdaddr_t *bdaddr)
 {
-	struct dev_info *dev = &devs[index];
-	disconnect_cp cp;
+	DBG("hci%d", index);
 
-	DBG("hci%d handle %u", index, handle);
-
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(handle);
-	cp.reason = HCI_OE_USER_ENDED_CONNECTION;
-
-	if (hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_DISCONNECT,
-						DISCONNECT_CP_SIZE, &cp) < 0)
-		return -errno;
-
-	return 0;
+	return disconnect_addr(index, bdaddr, HCI_OE_USER_ENDED_CONNECTION);
 }
 
 static int hciops_remove_bonding(int index, bdaddr_t *bdaddr)
@@ -2860,23 +3190,6 @@ static int hciops_remove_bonding(int index, bdaddr_t *bdaddr)
 	return 0;
 }
 
-static int hciops_request_authentication(int index, uint16_t handle)
-{
-	struct dev_info *dev = &devs[index];
-	auth_requested_cp cp;
-
-	DBG("hci%d handle %u", index, handle);
-
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(handle);
-
-	if (hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_AUTH_REQUESTED,
-					AUTH_REQUESTED_CP_SIZE, &cp) < 0)
-		return -errno;
-
-	return 0;
-}
-
 static int hciops_pincode_reply(int index, bdaddr_t *bdaddr, const char *pin)
 {
 	struct dev_info *dev = &devs[index];
@@ -2966,27 +3279,6 @@ static int hciops_passkey_reply(int index, bdaddr_t *bdaddr, uint32_t passkey)
 	return err;
 }
 
-static int hciops_get_auth_info(int index, bdaddr_t *bdaddr, uint8_t *auth)
-{
-	struct dev_info *dev = &devs[index];
-	struct hci_auth_info_req req;
-	char addr[18];
-
-	ba2str(bdaddr, addr);
-	DBG("hci%d dba %s", index, addr);
-
-	memset(&req, 0, sizeof(req));
-	bacpy(&req.bdaddr, bdaddr);
-
-	if (ioctl(dev->sk, HCIGETAUTHINFO, (unsigned long) &req) < 0)
-		return -errno;
-
-	if (auth)
-		*auth = req.type;
-
-	return 0;
-}
-
 static int hciops_read_scan_enable(int index)
 {
 	struct dev_info *dev = &devs[index];
@@ -3021,48 +3313,6 @@ static int hciops_enable_le(int index)
 	return 0;
 }
 
-static uint8_t get_uuid_mask(uint16_t uuid16)
-{
-	switch (uuid16) {
-	case DIALUP_NET_SVCLASS_ID:
-	case CIP_SVCLASS_ID:
-		return 0x42;	/* Telephony & Networking */
-	case IRMC_SYNC_SVCLASS_ID:
-	case OBEX_OBJPUSH_SVCLASS_ID:
-	case OBEX_FILETRANS_SVCLASS_ID:
-	case IRMC_SYNC_CMD_SVCLASS_ID:
-	case PBAP_PSE_SVCLASS_ID:
-		return 0x10;	/* Object Transfer */
-	case HEADSET_SVCLASS_ID:
-	case HANDSFREE_SVCLASS_ID:
-		return 0x20;	/* Audio */
-	case CORDLESS_TELEPHONY_SVCLASS_ID:
-	case INTERCOM_SVCLASS_ID:
-	case FAX_SVCLASS_ID:
-	case SAP_SVCLASS_ID:
-	/*
-	 * Setting the telephony bit for the handsfree audio gateway
-	 * role is not required by the HFP specification, but the
-	 * Nokia 616 carkit is just plain broken! It will refuse
-	 * pairing without this bit set.
-	 */
-	case HANDSFREE_AGW_SVCLASS_ID:
-		return 0x40;	/* Telephony */
-	case AUDIO_SOURCE_SVCLASS_ID:
-	case VIDEO_SOURCE_SVCLASS_ID:
-		return 0x08;	/* Capturing */
-	case AUDIO_SINK_SVCLASS_ID:
-	case VIDEO_SINK_SVCLASS_ID:
-		return 0x04;	/* Rendering */
-	case PANU_SVCLASS_ID:
-	case NAP_SVCLASS_ID:
-	case GN_SVCLASS_ID:
-		return 0x02;	/* Networking */
-	default:
-		return 0;
-	}
-}
-
 static uint8_t generate_service_class(int index)
 {
 	struct dev_info *dev = &devs[index];
@@ -3070,12 +3320,9 @@ static uint8_t generate_service_class(int index)
 	uint8_t val = 0;
 
 	for (l = dev->uuids; l != NULL; l = g_slist_next(l)) {
-		uuid_t *uuid = l->data;
+		struct uuid_info *uuid = l->data;
 
-		if (uuid->type != SDP_UUID16)
-			continue;
-
-		val |= get_uuid_mask(uuid->value.uuid16);
+		val |= uuid->svc_hint;
 	}
 
 	return val;
@@ -3117,13 +3364,18 @@ static int update_service_classes(int index)
 	return err;
 }
 
-static int hciops_add_uuid(int index, uuid_t *uuid)
+static int hciops_add_uuid(int index, uuid_t *uuid, uint8_t svc_hint)
 {
 	struct dev_info *dev = &devs[index];
+	struct uuid_info *info;
 
 	DBG("hci%d", index);
 
-	dev->uuids = g_slist_append(dev->uuids, g_memdup(uuid, sizeof(*uuid)));
+	info = g_new0(struct uuid_info, 1);
+	memcpy(&info->uuid, uuid, sizeof(*uuid));
+	info->svc_hint = svc_hint;
+
+	dev->uuids = g_slist_append(dev->uuids, info);
 
 	return update_service_classes(index);
 }
@@ -3194,6 +3446,119 @@ static int hciops_load_keys(int index, GSList *keys, gboolean debug_keys)
 	return 0;
 }
 
+static int hciops_set_io_capability(int index, uint8_t io_capability)
+{
+	struct dev_info *dev = &devs[index];
+
+	dev->io_capability = io_capability;
+
+	return 0;
+}
+
+static int request_authentication(int index, bdaddr_t *bdaddr)
+{
+	struct dev_info *dev = &devs[index];
+	auth_requested_cp cp;
+	uint16_t handle;
+	int err;
+
+	DBG("hci%d", index);
+
+	err = get_handle(index, bdaddr, &handle);
+	if (err < 0)
+		return err;
+
+	memset(&cp, 0, sizeof(cp));
+	cp.handle = htobs(handle);
+
+	if (hci_send_cmd(dev->sk, OGF_LINK_CTL, OCF_AUTH_REQUESTED,
+					AUTH_REQUESTED_CP_SIZE, &cp) < 0)
+		return -errno;
+
+	return 0;
+}
+
+static void bonding_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
+{
+	struct bt_conn *conn = user_data;
+	struct dev_info *dev = conn->dev;
+
+	if (!conn->io) {
+		if (!err)
+			g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (err)
+		/* Wait proper error to be propagated by bonding complete */
+		return;
+
+	if (request_authentication(dev->id, &conn->bdaddr) < 0)
+		goto failed;
+
+	return;
+
+failed:
+	bonding_complete(dev, conn, HCI_UNSPECIFIED_ERROR);
+}
+
+static int hciops_create_bonding(int index, bdaddr_t *bdaddr, uint8_t io_cap)
+{
+	struct dev_info *dev = &devs[index];
+	BtIOSecLevel sec_level;
+	struct bt_conn *conn;
+	GError *err = NULL;
+
+	conn = get_connection(dev, bdaddr);
+
+	if (conn->io != NULL)
+		return -EBUSY;
+
+	conn->loc_cap = io_cap;
+
+	/* If our IO capability is NoInputNoOutput use medium security
+	 * level (i.e. don't require MITM protection) else use high
+	 * security level */
+	if (io_cap == 0x03)
+		sec_level = BT_IO_SEC_MEDIUM;
+	else
+		sec_level = BT_IO_SEC_HIGH;
+
+	conn->io = bt_io_connect(BT_IO_L2RAW, bonding_connect_cb, conn,
+					NULL, &err,
+					BT_IO_OPT_SOURCE_BDADDR, &dev->bdaddr,
+					BT_IO_OPT_DEST_BDADDR, bdaddr,
+					BT_IO_OPT_SEC_LEVEL, sec_level,
+					BT_IO_OPT_INVALID);
+	if (conn->io == NULL) {
+		error("bt_io_connect: %s", err->message);
+		g_error_free(err);
+		return -EIO;
+	}
+
+	conn->bonding_initiator = TRUE;
+
+	return 0;
+}
+
+static int hciops_cancel_bonding(int index, bdaddr_t *bdaddr)
+{
+	struct dev_info *dev = &devs[index];
+	struct bt_conn *conn;
+
+	DBG("hci%d", index);
+
+	conn = find_connection(dev, bdaddr);
+	if (conn == NULL || conn->io == NULL)
+		return -ENOTCONN;
+
+	g_io_channel_shutdown(conn->io, TRUE, NULL);
+	g_io_channel_unref(conn->io);
+	conn->io = NULL;
+
+	return 0;
+}
+
 static struct btd_adapter_ops hci_ops = {
 	.setup = hciops_setup,
 	.cleanup = hciops_cleanup,
@@ -3219,11 +3584,9 @@ static struct btd_adapter_ops hci_ops = {
 	.read_local_features = hciops_read_local_features,
 	.disconnect = hciops_disconnect,
 	.remove_bonding = hciops_remove_bonding,
-	.request_authentication = hciops_request_authentication,
 	.pincode_reply = hciops_pincode_reply,
 	.confirm_reply = hciops_confirm_reply,
 	.passkey_reply = hciops_passkey_reply,
-	.get_auth_info = hciops_get_auth_info,
 	.read_scan_enable = hciops_read_scan_enable,
 	.enable_le = hciops_enable_le,
 	.encrypt_link = hciops_encrypt_link,
@@ -3233,6 +3596,9 @@ static struct btd_adapter_ops hci_ops = {
 	.disable_cod_cache = hciops_disable_cod_cache,
 	.restore_powered = hciops_restore_powered,
 	.load_keys = hciops_load_keys,
+	.set_io_capability = hciops_set_io_capability,
+	.create_bonding = hciops_create_bonding,
+	.cancel_bonding = hciops_cancel_bonding,
 };
 
 static int hciops_init(void)
