@@ -38,28 +38,12 @@
 #include <glib.h>
 
 #include "btio.h"
-#include "gattrib.h"
-#include "att.h"
-#include "gatt.h"
 #include "sdpd.h"
 #include "glib-helper.h"
 
 /* Number of seconds to keep a sdp_session_t in the cache */
 #define CACHE_TIMEOUT 2
 
-struct gattrib_context {
-	bdaddr_t src;
-	bdaddr_t dst;
-	GAttrib *attrib;
-	GIOChannel *io;
-	bt_primary_t cb;
-	bt_destroy_t destroy;
-	gpointer user_data;
-	GSList *uuids;
-};
-
-static GSList *gattrib_list = NULL;
-
 struct cached_sdp_session {
 	bdaddr_t src;
 	bdaddr_t dst;
@@ -69,23 +53,6 @@ struct cached_sdp_session {
 
 static GSList *cached_sdp_sessions = NULL;
 
-static void gattrib_context_free(struct gattrib_context *ctxt)
-{
-	gattrib_list = g_slist_remove(gattrib_list, ctxt);
-	if (ctxt->destroy)
-		ctxt->destroy(ctxt->user_data);
-
-	g_slist_foreach(ctxt->uuids, (GFunc) g_free, NULL);
-	g_slist_free(ctxt->uuids);
-	g_attrib_unref(ctxt->attrib);
-	if (ctxt->io) {
-		g_io_channel_unref(ctxt->io);
-		g_io_channel_shutdown(ctxt->io, FALSE, NULL);
-	}
-
-	g_free(ctxt);
-}
-
 static gboolean cached_session_expired(gpointer user_data)
 {
 	struct cached_sdp_session *cached = user_data;
@@ -363,16 +330,6 @@ int bt_search_service(const bdaddr_t *src, const bdaddr_t *dst,
 	return 0;
 }
 
-int bt_discover_services(const bdaddr_t *src, const bdaddr_t *dst,
-		bt_callback_t cb, void *user_data, bt_destroy_t destroy)
-{
-	uuid_t uuid;
-
-	sdp_uuid16_create(&uuid, PUBLIC_BROWSE_GROUP);
-
-	return bt_search_service(src, dst, &uuid, cb, user_data, destroy);
-}
-
 static gint find_by_bdaddr(gconstpointer data, gconstpointer user_data)
 {
 	const struct search_context *ctxt = data, *search = user_data;
@@ -381,16 +338,22 @@ static gint find_by_bdaddr(gconstpointer data, gconstpointer user_data)
 					bacmp(&ctxt->src, &search->src));
 }
 
-static gint gattrib_find_by_bdaddr(gconstpointer data, gconstpointer user_data)
+int bt_cancel_discovery(const bdaddr_t *src, const bdaddr_t *dst)
 {
-	const struct gattrib_context *ctxt = data, *search = user_data;
+	struct search_context match, *ctxt;
+	GSList *l;
 
-	return (bacmp(&ctxt->dst, &search->dst) &&
-					bacmp(&ctxt->src, &search->src));
-}
+	memset(&match, 0, sizeof(match));
+	bacpy(&match.src, src);
+	bacpy(&match.dst, dst);
+
+	/* Ongoing SDP Discovery */
+	l = g_slist_find_custom(context_list, &match, find_by_bdaddr);
+	if (l == NULL)
+		return -ENOENT;
+
+	ctxt = l->data;
 
-static int cancel_sdp(struct search_context *ctxt)
-{
 	if (!ctxt->session)
 		return -ENOTCONN;
 
@@ -405,164 +368,6 @@ static int cancel_sdp(struct search_context *ctxt)
 	return 0;
 }
 
-static int cancel_gattrib(struct gattrib_context *ctxt)
-{
-	if (ctxt->attrib)
-		g_attrib_cancel_all(ctxt->attrib);
-
-	gattrib_context_free(ctxt);
-
-	return 0;
-}
-
-int bt_cancel_discovery(const bdaddr_t *src, const bdaddr_t *dst)
-{
-	struct search_context sdp_ctxt;
-	struct gattrib_context gatt_ctxt;
-	GSList *match;
-
-	memset(&sdp_ctxt, 0, sizeof(sdp_ctxt));
-	bacpy(&sdp_ctxt.src, src);
-	bacpy(&sdp_ctxt.dst, dst);
-
-	/* Ongoing SDP Discovery */
-	match = g_slist_find_custom(context_list, &sdp_ctxt, find_by_bdaddr);
-	if (match)
-		return cancel_sdp(match->data);
-
-	memset(&gatt_ctxt, 0, sizeof(gatt_ctxt));
-	bacpy(&gatt_ctxt.src, src);
-	bacpy(&gatt_ctxt.dst, dst);
-
-	/* Ongoing Discover All Primary Services */
-	match = g_slist_find_custom(gattrib_list, &gatt_ctxt,
-						gattrib_find_by_bdaddr);
-	if (match == NULL)
-		return -ENOTCONN;
-
-	return cancel_gattrib(match->data);
-}
-
-static void primary_cb(guint8 status, const guint8 *pdu, guint16 plen,
-							gpointer user_data)
-{
-	struct gattrib_context *ctxt = user_data;
-	struct att_data_list *list;
-	unsigned int i, err;
-	uint16_t end;
-
-	if (status == ATT_ECODE_ATTR_NOT_FOUND) {
-		err = 0;
-		goto done;
-	}
-
-	if (status != 0) {
-		err = -EIO;
-		goto done;
-	}
-
-	list = dec_read_by_grp_resp(pdu, plen);
-	if (list == NULL) {
-		err = -EPROTO;
-		goto done;
-	}
-
-	for (i = 0, end = 0; i < list->num; i++) {
-		const uint8_t *data = list->data[i];
-		char *prim;
-		uuid_t u128, u16;
-
-		end = att_get_u16(&data[2]);
-
-		if (list->len == 6) {
-			sdp_uuid16_create(&u16,
-					att_get_u16(&data[4]));
-			sdp_uuid16_to_uuid128(&u128, &u16);
-
-		} else if (list->len == 20)
-			sdp_uuid128_create(&u128, &data[4]);
-		else
-			/* Skipping invalid data */
-			continue;
-
-		prim = bt_uuid2string(&u128);
-		ctxt->uuids = g_slist_append(ctxt->uuids, prim);
-	}
-
-	att_data_list_free(list);
-	err = 0;
-
-	if (end != 0xffff) {
-		gatt_discover_primary(ctxt->attrib, end + 1, 0xffff, NULL,
-							primary_cb, ctxt);
-		return;
-	}
-
-done:
-	ctxt->cb(ctxt->uuids, err, ctxt->user_data);
-	gattrib_context_free(ctxt);
-}
-
-static void connect_cb(GIOChannel *io, GError *gerr, gpointer user_data)
-{
-	struct gattrib_context *ctxt = user_data;
-
-	if (gerr) {
-		ctxt->cb(NULL, -EIO, ctxt->user_data);
-		gattrib_context_free(ctxt);
-		return;
-	}
-
-	ctxt->attrib = g_attrib_new(io);
-	gatt_discover_primary(ctxt->attrib, 0x0001, 0xffff, NULL, primary_cb,
-									ctxt);
-}
-
-int bt_discover_primary(const bdaddr_t *src, const bdaddr_t *dst, int psm,
-					bt_primary_t cb, void *user_data,
-					bt_destroy_t destroy)
-{
-	struct gattrib_context *ctxt;
-	GIOChannel *io;
-	GError *gerr = NULL;
-
-	ctxt = g_try_new0(struct gattrib_context, 1);
-	if (ctxt == NULL)
-		return -ENOMEM;
-
-	bacpy(&ctxt->src, src);
-	bacpy(&ctxt->dst, dst);
-	ctxt->user_data = user_data;
-	ctxt->cb = cb;
-	ctxt->destroy = destroy;
-
-	if (psm < 0)
-		io = bt_io_connect(BT_IO_L2CAP, connect_cb, ctxt, NULL, &gerr,
-				BT_IO_OPT_SOURCE_BDADDR, src,
-				BT_IO_OPT_DEST_BDADDR, dst,
-				BT_IO_OPT_CID, GATT_CID,
-				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
-				BT_IO_OPT_INVALID);
-	else
-		io = bt_io_connect(BT_IO_L2CAP, connect_cb, ctxt, NULL, &gerr,
-				BT_IO_OPT_SOURCE_BDADDR, src,
-				BT_IO_OPT_DEST_BDADDR, dst,
-				BT_IO_OPT_PSM, psm,
-				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
-				BT_IO_OPT_INVALID);
-
-	if (io == NULL) {
-		gattrib_context_free(ctxt);
-		return -EIO;
-	}
-
-	ctxt->io = io;
-
-	gattrib_list = g_slist_append(gattrib_list, ctxt);
-
-	return 0;
-}
-
 char *bt_uuid2string(uuid_t *uuid)
 {
 	gchar *str;
@@ -780,25 +585,3 @@ GSList *bt_string2list(const gchar *str)
 
 	return l;
 }
-
-char *bt_extract_eir_name(uint8_t *data, uint8_t *type)
-{
-	if (!data || !type)
-		return NULL;
-
-	if (data[0] == 0)
-		return NULL;
-
-	if (type)
-		*type = data[1];
-
-	switch (data[1]) {
-	case EIR_NAME_SHORT:
-	case EIR_NAME_COMPLETE:
-		if (!g_utf8_validate((char *) (data + 2), data[0] - 1, NULL))
-			return g_strdup("");
-		return g_strndup((char *) (data + 2), data[0] - 1);
-	}
-
-	return NULL;
-}
