@@ -59,25 +59,6 @@ const char *manager_get_base_path(void)
 	return base_path;
 }
 
-void manager_update_svc(struct btd_adapter* adapter, uint8_t svc)
-{
-	adapter_set_service_classes(adapter, svc);
-}
-
-static inline DBusMessage *invalid_args(DBusMessage *msg)
-{
-	return g_dbus_create_error(msg,
-			ERROR_INTERFACE ".InvalidArguments",
-			"Invalid arguments in method call");
-}
-
-static inline DBusMessage *no_such_adapter(DBusMessage *msg)
-{
-	return g_dbus_create_error(msg,
-			ERROR_INTERFACE ".NoSuchAdapter",
-			"No such adapter");
-}
-
 static DBusMessage *default_adapter(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -87,7 +68,7 @@ static DBusMessage *default_adapter(DBusConnection *conn,
 
 	adapter = manager_find_adapter_by_id(default_adapter_id);
 	if (!adapter)
-		return no_such_adapter(msg);
+		return btd_error_no_such_adapter(msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -120,15 +101,18 @@ static DBusMessage *find_adapter(DBusConnection *conn,
 		path = adapter_any_get_path();
 		if (path != NULL)
 			goto done;
-		return no_such_adapter(msg);
+		return btd_error_no_such_adapter(msg);
 	} else if (!strncmp(pattern, "hci", 3) && strlen(pattern) >= 4) {
 		dev_id = atoi(pattern + 3);
 		adapter = manager_find_adapter_by_id(dev_id);
-	} else
-		adapter = manager_find_adapter_by_address(pattern);
+	} else {
+		bdaddr_t bdaddr;
+		str2ba(pattern, &bdaddr);
+		adapter = manager_find_adapter(&bdaddr);
+	}
 
 	if (!adapter)
-		return no_such_adapter(msg);
+		return btd_error_no_such_adapter(msg);
 
 	path = adapter_get_path(adapter);
 
@@ -198,9 +182,6 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	for (i = 0, list = adapters; list; list = list->next) {
 		struct btd_adapter *adapter = list->data;
 
-		if (!adapter_is_ready(adapter))
-			continue;
-
 		array[i] = (char *) adapter_get_path(adapter);
 		i++;
 	}
@@ -250,9 +231,6 @@ static void manager_update_adapters(void)
 	for (i = 0, list = adapters; list; list = list->next) {
 		struct btd_adapter *adapter = list->data;
 
-		if (!adapter_is_ready(adapter))
-			continue;
-
 		array[i] = (char *) adapter_get_path(adapter);
 		i++;
 	}
@@ -264,6 +242,26 @@ static void manager_update_adapters(void)
 	g_free(array);
 }
 
+static void manager_set_default_adapter(int id)
+{
+	struct btd_adapter *adapter;
+	const gchar *path;
+
+	default_adapter_id = id;
+
+	adapter = manager_find_adapter_by_id(id);
+	if (!adapter)
+		return;
+
+	path = adapter_get_path(adapter);
+
+	g_dbus_emit_signal(connection, "/",
+			MANAGER_INTERFACE,
+			"DefaultAdapterChanged",
+			DBUS_TYPE_OBJECT_PATH, &path,
+			DBUS_TYPE_INVALID);
+}
+
 static void manager_remove_adapter(struct btd_adapter *adapter)
 {
 	uint16_t dev_id = adapter_get_dev_id(adapter);
@@ -307,15 +305,6 @@ static gint adapter_id_cmp(gconstpointer a, gconstpointer b)
 	return dev_id == id ? 0 : -1;
 }
 
-static gint adapter_path_cmp(gconstpointer a, gconstpointer b)
-{
-	struct btd_adapter *adapter = (struct btd_adapter *) a;
-	const char *path = b;
-	const gchar *adapter_path = adapter_get_path(adapter);
-
-	return strcmp(adapter_path, path);
-}
-
 static gint adapter_cmp(gconstpointer a, gconstpointer b)
 {
 	struct btd_adapter *adapter = (struct btd_adapter *) a;
@@ -327,19 +316,6 @@ static gint adapter_cmp(gconstpointer a, gconstpointer b)
 	return bacmp(&src, bdaddr);
 }
 
-static gint adapter_address_cmp(gconstpointer a, gconstpointer b)
-{
-	struct btd_adapter *adapter = (struct btd_adapter *) a;
-	const char *address = b;
-	bdaddr_t bdaddr;
-	char addr[18];
-
-	adapter_get_address(adapter, &bdaddr);
-	ba2str(&bdaddr, addr);
-
-	return strcasecmp(addr, address);
-}
-
 struct btd_adapter *manager_find_adapter(const bdaddr_t *sba)
 {
 	GSList *match;
@@ -351,28 +327,6 @@ struct btd_adapter *manager_find_adapter(const bdaddr_t *sba)
 	return match->data;
 }
 
-struct btd_adapter *manager_find_adapter_by_address(const char *address)
-{
-	GSList *match;
-
-	match = g_slist_find_custom(adapters, address, adapter_address_cmp);
-	if (!match)
-		return NULL;
-
-	return match->data;
-}
-
-struct btd_adapter *manager_find_adapter_by_path(const char *path)
-{
-	GSList *match;
-
-	match = g_slist_find_custom(adapters, path, adapter_path_cmp);
-	if (!match)
-		return NULL;
-
-	return match->data;
-}
-
 struct btd_adapter *manager_find_adapter_by_id(int id)
 {
 	GSList *match;
@@ -385,6 +339,11 @@ struct btd_adapter *manager_find_adapter_by_id(int id)
 	return match->data;
 }
 
+void manager_foreach_adapter(adapter_cb func, gpointer user_data)
+{
+	g_slist_foreach(adapters, (GFunc) func, user_data);
+}
+
 GSList *manager_get_adapters(void)
 {
 	return adapters;
@@ -402,26 +361,47 @@ void manager_add_adapter(const char *path)
 	btd_stop_exit_timer();
 }
 
-int manager_register_adapter(int id, gboolean devup)
+struct btd_adapter *btd_manager_register_adapter(int id)
 {
 	struct btd_adapter *adapter;
+	const char *path;
 
 	adapter = manager_find_adapter_by_id(id);
 	if (adapter) {
 		error("Unable to register adapter: hci%d already exist", id);
-		return -1;
+		return NULL;
 	}
 
-	adapter = adapter_create(connection, id, devup);
+	adapter = adapter_create(connection, id);
 	if (!adapter)
-		return -1;
+		return NULL;
 
 	adapters = g_slist_append(adapters, adapter);
 
-	return 0;
+	if (!adapter_init(adapter)) {
+		btd_adapter_unref(adapter);
+		return NULL;
+	}
+
+	path = adapter_get_path(adapter);
+	g_dbus_emit_signal(connection, "/",
+				MANAGER_INTERFACE, "AdapterAdded",
+				DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID);
+
+	manager_update_adapters();
+
+	btd_stop_exit_timer();
+
+	if (default_adapter_id < 0)
+		manager_set_default_adapter(id);
+
+	DBG("Adapter %s registered", path);
+
+	return btd_adapter_ref(adapter);
 }
 
-int manager_unregister_adapter(int id)
+int btd_manager_unregister_adapter(int id)
 {
 	struct btd_adapter *adapter;
 	const gchar *path;
@@ -439,75 +419,13 @@ int manager_unregister_adapter(int id)
 	return 0;
 }
 
-int manager_start_adapter(int id)
-{
-	struct btd_adapter *adapter;
-	int ret;
-
-	adapter = manager_find_adapter_by_id(id);
-	if (!adapter) {
-		error("Getting device data failed: hci%d", id);
-		return -EINVAL;
-	}
-
-	ret = adapter_start(adapter);
-	if (ret < 0)
-		return ret;
-
-	if (default_adapter_id < 0)
-		manager_set_default_adapter(id);
-
-	return ret;
-}
-
-int manager_stop_adapter(int id)
-{
-	struct btd_adapter *adapter;
-
-	adapter = manager_find_adapter_by_id(id);
-	if (!adapter) {
-		error("Getting device data failed: hci%d", id);
-		return -EINVAL;
-	}
-
-	return adapter_stop(adapter);
-}
-
-int manager_get_default_adapter()
-{
-	return default_adapter_id;
-}
-
-void manager_set_default_adapter(int id)
-{
-	struct btd_adapter *adapter;
-	const gchar *path;
-
-	default_adapter_id = id;
-
-	adapter = manager_find_adapter_by_id(id);
-	if (!adapter)
-		return;
-
-	path = adapter_get_path(adapter);
-
-	g_dbus_emit_signal(connection, "/",
-			MANAGER_INTERFACE,
-			"DefaultAdapterChanged",
-			DBUS_TYPE_OBJECT_PATH, &path,
-			DBUS_TYPE_INVALID);
-}
-
-void btd_manager_set_offline(gboolean offline)
+void btd_manager_set_did(uint16_t vendor, uint16_t product, uint16_t version)
 {
 	GSList *l;
 
 	for (l = adapters; l != NULL; l = g_slist_next(l)) {
 		struct btd_adapter *adapter = l->data;
 
-		if (offline)
-			btd_adapter_switch_offline(adapter);
-		else
-			btd_adapter_restore_powered(adapter);
+		btd_adapter_set_did(adapter, vendor, product, version);
 	}
 }
